#!/usr/bin/env python
# coding=utf-8
from __future__ import absolute_import, print_function

import argparse
import base64
import hashlib
import json
import os
import re
import sys
import time

import redis
import six
import yaml
from trmaid.plugins import plugins
from trmaid.rpc_client import rpc_client

__version__ = "0.6"

if sys.version_info[0] < 3:
    from urlparse import urlparse
    string_types = unicode
else:
    from urllib.parse import urlparse
    string_types = str


CONF_FILE = '~/.trmaid.yaml'
CONF_FILE_FULL = os.path.expanduser(CONF_FILE)


def msg_and_exit(*msgs):
    for msg in msgs:
        sys.stdout.write('%s\n' % msg)
        sys.stdout.flush()
    sys.exit(1)


def textme(elem, encoding='utf-8'):
    if isinstance(elem, dict):
        for key in list(elem.keys()):
            elem[key] = textme(elem[key])
    elif isinstance(elem, (list, tuple)):
        res = []
        for e in elem:
            res.append(textme(e))
        if isinstance(elem, tuple):
            elem = tuple(res)
        else:
            elem = res
    if isinstance(elem, six.binary_type):
        elem = elem.decode(encoding)
    return elem


def iterdirs(startpath):
    directory = os.path.normpath(startpath).split(os.path.sep)
    while directory:
        yield os.path.sep.join(directory)
        directory.pop()


def human_size(num):
    sizes = ['B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB']
    while num > 1024:
        num /= 1024
        sizes.pop(0)
    return '%.2f%s' % (num, sizes[0])


class TRMaid(object):

    server = None

    def __init__(self, config_file=CONF_FILE_FULL, verbose=0, server=None):
        config_file = os.path.expanduser(config_file)
        if not os.path.exists(config_file):
            msg_and_exit(
                'Please create config file called %s.' % config_file,
                'Go see the example at:',
                '  https://github.com/tanelpuhu/trmaid',
            )

        self.verbose = verbose

        # config
        self.msg('using config from %s' % config_file)
        self.conf = textme(yaml.load(open(config_file)))

        if server and server not in self.conf:
            msg_and_exit('Server "%s" is not configured!' % server)
        elif server:
            self.msg('selecting server %s' % server)
            self.conf = self.conf[server]
            self.server = server

        if 'transmission' not in self.conf:
            items = list(self.conf.items())
            for name, value in items:
                if len(items) == 1 or value.get('general', {}).get('default'):
                    self.msg('selecting server %s' % name)
                    self.conf = value
                    self.server = name
                    break

            if 'transmission' not in self.conf:
                msg_and_exit(
                    'Something is fishy with your configuration! '
                    'No default server?'
                )

        general = self.conf.get('general', {})
        transmission = self.conf.get('transmission', {})

        # verbose
        if not self.verbose and general.get('verbose', False):
            self.verbose = 1
        elif not self.verbose:
            self.verbose = 0

        # redis
        if 'redis' in general:
            rds_url = general['redis']
        else:
            rds_url = 'redis://localhost:6379'

        try:
            self.rds = redis.from_url(rds_url)
            self.rds.ping()
        except redis.exceptions.ConnectionError:
            msg_and_exit('Please install and run redis on %s' % rds_url)

        # transmission conf check
        if 'url' not in transmission:
            msg_and_exit('transmission.url not configured')

        # rpc stuff
        verify = general.get('verify', True)

        self.rpc = rpc_client(
            url=transmission['url'],
            username=transmission.get('username'),
            password=transmission.get('password'),
            verify=verify,
            textme=textme,
        )

    def msg(self, *msgs, **kwargs):
        level = kwargs.get('level', 1)
        indent = 0
        if 'indent' in kwargs:
            indent = kwargs.pop('indent')

        if self.verbose >= level:
            for msg in msgs:
                if indent:
                    sys.stdout.write(' ' * indent)
                sys.stdout.write(' => ')
                sys.stdout.write('%s\n' % msg)
                sys.stdout.flush()

    def verbose_func(self, func, *args):
        if self.verbose:
            sys.stdout.write('%s(%s) => ' % (
                func.__name__,
                ', '.join([json.dumps(x, indent=2) for x in args]),
            ))
            sys.stdout.flush()
        result = func(*args)
        if self.verbose:
            sys.stdout.write('%s\n' % json.dumps(
                result, indent=2, sort_keys=True
            ))
            sys.stdout.flush()
        return result

    def collect_stats(self):
        # is stats configured
        reset_day = self.conf.get('general', {}).get('stats')
        if not reset_day:
            return

        try:
            reset_day = int(float(reset_day))
        except ValueError:
            msg_and_exit('general.stats not number!')
            return

        # fetch session stats
        args = self.rpc.session_stats().get('arguments', {})
        if not args:
            return

        key = '%s-%s' % (time.strftime(
            '%Y.%m', time.localtime(time.time() - ((reset_day - 1) * 3600 * 24))
        ), reset_day)

        uploadedBytes = args['cumulative-stats']['uploadedBytes']
        downloadedBytes = args['cumulative-stats']['downloadedBytes']

        # setup keys
        upload_start_key = self.rdskey('stats', key, 'upload', 'start')
        upload_end_key = self.rdskey('stats', key, 'upload', 'end')
        download_start_key = self.rdskey('stats', key, 'download', 'start')
        download_end_key = self.rdskey('stats', key, 'download', 'end')

        pipe = self.rds.pipeline()

        # fetch saved stats
        pipe.get(upload_start_key).get(download_start_key)
        upload_start, download_start = pipe.execute()

        # if not yet saved, save uload for this 'key'... as start
        if upload_start is None:
            upload_start = uploadedBytes
            pipe.set(upload_start_key, upload_start)
        else:
            upload_start = float(upload_start)

        # and download too
        if download_start is None:
            download_start = downloadedBytes
            pipe.set(download_start_key, download_start)
        else:
            download_start = float(download_start)

        # save current values as end
        pipe.set(download_end_key, downloadedBytes)
        pipe.set(upload_end_key, uploadedBytes)

        for key in [
            upload_start_key,
            upload_end_key,
            download_start_key,
            download_end_key,
        ]:
            pipe.expire(key, 3600 * 24 * 32)

        pipe.execute()

        # disaplay approximate upload and download
        self.msg('approximate upload this month: %s' % (
            human_size(float(uploadedBytes) - upload_start)
        ))
        self.msg('approximate download this month: %s' % (
            human_size(float(downloadedBytes) - download_start)
        ))

    def handle_session(self):
        if 'session' not in self.conf:
            return self.msg('No [session] section')

        self.msg('handling session...')
        req = {}
        args = self.rpc.session_get().get('arguments', {})

        free = args.get('download-dir-free-space', 0)
        if free:
            units = float(1024)
            free_gb = free / (units * units * units)
            if free_gb < 20:
                self.msg(
                    'ONLY %s FREE SPACE LEFT!!' % human_size(free).upper(),
                    level=0
                )
            else:
                self.msg('%s free space left' % human_size(free))

        keys = [
            # Booleans
            "alt-speed-enabled",
            "alt-speed-time-enabled",
            "blocklist-enabled",
            "dht-enabled",
            "download-queue-enabled",
            "idle-seeding-limit-enabled",
            "incomplete-dir-enabled",
            "lpd-enabled",
            "peer-port-random-on-start",
            "pex-enabled",
            "port-forwarding-enabled",
            "queue-stalled-enabled",
            "rename-partial-files",
            "script-torrent-done-enabled",
            "seed-queue-enabled",
            "seedRatioLimited",
            "speed-limit-down-enabled",
            "speed-limit-up-enabled",
            "start-added-torrents",
            "trash-original-torrent-files",
            "utp-enabled",
            # Doubles / Floats
            "seedRatioLimit",
            # Integers
            "alt-speed-down",
            "alt-speed-time-begin",
            "alt-speed-time-day",
            "alt-speed-time-end",
            "alt-speed-up",
            "cache-size-mb",
            "download-queue-size",
            "idle-seeding-limit",
            "peer-limit-global",
            "peer-limit-per-torrent",
            "peer-port",
            "queue-stalled-minutes",
            "seed-queue-size",
            "speed-limit-down",
            "speed-limit-up",
            # Text / Strings
            "blocklist-url",
            "download-dir",
            "encryption",
            "incomplete-dir",
            "script-torrent-done-filename",
        ]
        session = self.conf['session']
        for key in keys:
            value = session.get(key)
            if value is None:
                continue
            current = args.get(key)
            if current != value:
                req[key] = value
            else:
                self.msg(' > %s=%s' % (key, current), level=2)

        # blocklist
        blocklist_enabled = session.get('blocklist-enabled')
        blocklist_update_interval = session.get('blocklist-update-interval')
        blocklist_url = session.get('blocklist-url')

        if blocklist_enabled and blocklist_update_interval and blocklist_url:
            rdskey = self.rdskey('blocklist-update')
            if self.rds.setnx(rdskey, '1'):
                self.rds.expire(rdskey, blocklist_update_interval * 3600)
                self.verbose_func(self.rpc.blocklist_update)

        if req:
            self.verbose_func(self.rpc.session_set, req)

    def handle_all_torrents(self):
        torrents = self.rpc.torrent_get({'fields': [
            'bandwidthPriority',
            'doneDate',
            'downloadDir',
            'downloadLimit',
            'downloadLimited',
            'error',
            'errorString',
            'hashString',
            'honorsSessionLimits',
            'id',
            'isPrivate',
            'name',
            'peer-limit',
            'peer-limit-global',
            'peer-limit-per-torrent',
            'percentDone',
            'queuePosition',
            'seedIdleLimit',
            'seedIdleMode',
            'seedRatioLimit',
            'seedRatioMode',
            'status',
            'totalSize',
            'trackers',
            'uploadLimit',
            'uploadLimited',
            'uploadRatio',
        ]})
        if torrents.get('result') == 'success':
            torrents = torrents.get('arguments', {}).get('torrents', [])
            self.msg('there are %d total torrents...' % len(torrents))
            for tor in torrents:
                tor['status_text'] = self.rpc.get_status(tor['status'])
                tor['percentDone'] *= 100.0
                self.handle_torrent(tor)

    def handle_torrent(self, tor):
        if tor['error']:
            self.msg(' ! {id:<4} {name}: {errorString}'.format(**tor), indent=1)
        sections = self.conf.get('tracker', [])
        for tracker in tor['trackers']:
            announce = tracker.get('announce')
            if not announce:
                continue
            hostname = urlparse(announce).hostname
            if hostname.startswith('www.'):
                hostname = hostname[4:]
            if hostname not in sections:
                continue
            return self.handle_torrent_section(tor, 'tracker', hostname)

        sections = list(self.conf.get('location', {}).keys())
        for location in iterdirs(tor['downloadDir']):
            if location and location in sections:
                return self.handle_torrent_section(tor, 'location', location)

        general = self.conf.get('tracker', {}).get('general')
        if general:
            return self.handle_torrent_section(tor, 'tracker', 'general')

    def handle_torrent_section(self, tor, section, subsection):
        conf = self.conf.get(section, {}).get(subsection, {})
        for key in ['download-dir', 'location']:
            if key in conf:
                correct = os.path.normpath(conf.get(key))
                current = os.path.normpath(tor['downloadDir'])
                if current != correct:
                    self.verbose_func(
                        self.rpc.torrent_set_location, tor['id'], correct)
                    break

        if round(tor['percentDone'], 2) > 99.99:
            remove = []
            stopped = tor['status_text'] == 'Stopped'
            for hkey in ['remove-stopped-after-hours', 'remove-after-hours']:
                if hkey in conf:
                    if not stopped and hkey == 'remove-stopped-after-hours':
                        remove.append(0)
                        break
                    limit = float(conf[hkey])
                    has_been = time.time() - tor['doneDate']
                    if has_been > limit * 3600:
                        remove.append(1)
                    else:
                        remove.append(0)
                        break

            for rkey in ['remove-stopped-after-ratio', 'remove-after-ratio']:
                if rkey in conf:
                    if not stopped and rkey == 'remove-stopped-after-ratio':
                        remove.append(0)
                        break
                    limit = float(conf[rkey])
                    if tor['uploadRatio'] >= limit:
                        remove.append(1)
                    else:
                        remove.append(0)
                        break

            if remove and len(remove) == sum(remove):
                self.msg('removing %s...' % tor['name'])
                self.verbose_func(self.rpc.torrent_remove, tor['id'])
                return

        req = self.get_torrent_set_args(section, subsection, tor)
        if req:
            req['ids'] = tor['id']
            self.verbose_func(self.rpc.torrent_set, req)

    def get_torrent_add_args(self, section, subsection):
        req = {}
        keys = [
            # Booleans
            "paused",
            # Doubles / Floats
            "seedRatioLimit",
            # Integers
            "peer-limit",
            "bandwidthPriority",
            # Text / Strings
            "cookies",
            "download-dir",
        ]
        for key in keys:
            value = self.conf.get(section, {}).get(subsection, {}).get(key)
            if value is None:
                continue
            req[key] = value

        return req

    def get_torrent_set_args(self, section, subsection, tor=None):
        req = {}
        keys = [
            # Booleans
            "downloadLimited",
            "honorsSessionLimits",
            "uploadLimited",
            # Doubles / Floats
            "seedRatioLimit",
            # Integers
            "bandwidthPriority",
            "downloadLimit",
            "peer-limit",
            "queuePosition",
            "seedIdleLimit",
            "seedIdleMode",
            "seedRatioMode",
            "uploadLimit",
        ]
        conf = self.conf.get(section, {}).get(subsection, {})
        for key in keys:
            value = conf.get(key)
            if value is None:
                continue
            if tor is None or tor[key] != value:
                req[key] = value

        if conf.get('location'):
            location = conf['location']
            if tor is None or tor['downloadDir'] != location:
                req['location'] = location

        if tor:
            if 'trackerAdd' in conf:
                trackers = [x['announce'] for x in tor['trackers']]
                missing = set(conf['trackerAdd']) - set(trackers)
                if missing:
                    req['trackerAdd'] = list(missing)
            if tor['trackers'] and 'trackerRemove' in conf:
                remove = set()
                for tracker in tor['trackers']:
                    if tracker['announce'] in conf['trackerRemove']:
                        remove.add(tracker['id'])
                if remove:
                    req['trackerRemove'] = list(remove)
        return req

    @staticmethod
    def md5(text):
        return hashlib.md5(text.encode('utf-8')).hexdigest()

    def rdskey(self, key, *args):
        parts = ['trmaid']
        if self.server:
            parts.append(self.server)
        parts.append(key)
        parts.extend(args)
        return ':'.join(parts)

    def add_links(self, section, subsection, links, check_duplicates=False):
        if check_duplicates:
            rdskey = self.rdskey(
                'added', self.md5(subsection), urlparse(subsection).hostname
            )
            existing = textme(self.rds.lrange(rdskey, 0, -1))
        else:
            rdskey = None
            existing = []

        conf = self.conf[section][subsection]
        ids = []
        add_req = self.get_torrent_add_args(section, subsection)
        for title, link in links:
            xhash = self.md5(link)
            if xhash in existing:
                self.msg(' - existing: %s' % (title or link), level=3)
                continue
            elif 'matches' in conf:
                if not title:
                    continue
                for pattern in conf['matches']:
                    if re.match(pattern, title, re.I):
                        self.msg(' + %s matches "%s"' % (
                            title, pattern
                        ), level=2)
                        break
                else:
                    continue

            self.msg(' > adding: %s' % link, level=2)
            add_req['filename'] = link
            res = self.verbose_func(self.rpc.torrent_add, add_req)
            result = res.get('result', '')

            if rdskey and result in ['success', 'duplicate torrent']:
                self.rds.lpush(rdskey, xhash)

            arguments = res.get('arguments', {})
            if 'torrent-added' in arguments:
                ids.append(arguments['torrent-added']['id'])

        if ids:
            if rdskey:
                self.rds.ltrim(rdskey, 0, 1000)
            set_req = self.get_torrent_set_args(section, subsection)
            if set_req:
                set_req['ids'] = ids
                self.verbose_func(self.rpc.torrent_set, set_req)

    def get_torrent_files(self, path):
        if not os.path.isdir(path):
            self.msg('! not an folder: %s' % path)
        else:
            for file in os.listdir(path):
                if not file.endswith('.torrent'):
                    continue
                yield os.path.join(path, file)

    def handle_watch_folders(self):
        folders = self.conf.get('watch-folders')
        if not folders:
            return
        for folder in list(folders.keys()):
            self.msg('handling folder %s...' % folder)
            ids = []
            add_req = self.get_torrent_add_args('watch-folders', folder)
            for file in self.get_torrent_files(folder):
                content = base64.b64encode(open(file).read())
                add_req['metainfo'] = content
                res = self.verbose_func(self.rpc.torrent_add, add_req)
                result = res.get('result', '')

                if result in ['success', 'duplicate torrent']:
                    os.rename(file, file + '.added')

                if result != 'success':
                    continue

                arguments = res.get('arguments', {})

                if arguments.get('torrent-added', None):
                    ids.append(arguments['torrent-added']['id'])

            if ids:
                set_req = self.get_torrent_set_args('watch-folders', folder)
                if set_req:
                    set_req['ids'] = ids
                    self.verbose_func(self.rpc.torrent_set, set_req)

    def handle_plugins(self):
        for section, cls in list(plugins.items()):
            for subsection, conf in list(self.conf.get(section, {}).items()):
                self.msg('handling %s %s...' % (section, subsection[:40]))
                plugin = cls(
                    key=subsection,
                    rds=self.rds,
                    textme=textme,
                )
                self.add_links(
                    section=section,
                    subsection=subsection,
                    links=plugin.get_links(),
                    check_duplicates=plugin.check_duplicates
                )

    def run(self):
        self.handle_session()
        self.collect_stats()
        self.handle_plugins()
        self.handle_watch_folders()
        self.handle_all_torrents()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-v',
        help='verbose',
        action='count',
        default=0,
        dest='verbose',
    )
    parser.add_argument(
        '-s',
        help='server name, use when there are multiple servers in one file',
        dest='server',
    )
    parser.add_argument(
        'config_file',
        help='specify config file other than %s' % CONF_FILE,
        nargs='?',
        default=CONF_FILE_FULL
    )
    parser.add_argument(
        '-V',
        help='Show version and exit',
        action='store_const',
        const=True,
        dest='version',
        default=False,
    )
    args = parser.parse_args()

    if args.version:
        sys.stdout.write('trmaid %s\n' % __version__)
        sys.stdout.flush()
        return

    TRMaid(
        config_file=args.config_file,
        verbose=args.verbose,
        server=args.server,
    ).run()

if __name__ == '__main__':
    main()
